"""Integration tests for the UncompressCommand."""

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a UncompressCommand instance."""
    yield pebble_shell.commands.UncompressCommand(shell=shell)


def test_name(command: pebble_shell.commands.UncompressCommand):
    assert command.name == "uncompress"


def test_category(command: pebble_shell.commands.UncompressCommand):
    assert command.category == "Data Processing"


def test_help(command: pebble_shell.commands.UncompressCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "uncompress" in output
    assert "Decompress files" in output
    assert "-f" in output
    assert "-v" in output


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "uncompress" in output
    assert "Decompress files" in output


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # uncompress with no args should fail and show usage
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should fail with usage message
    assert result == 1
    output = capture.get()
    assert any(
        msg in output
        for msg in ["usage", "filename required", "no files specified", "uncompress"]
    )


def test_execute_nonexistent_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with nonexistent file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/nonexistent/file.Z"])

    # Should fail with file not found error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["No such file", "not found", "error"])


def test_execute_regular_file_without_extension(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with regular file that doesn't have .Z extension
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should either fail with extension error or attempt decompression
    if result == 1:
        output = capture.get()
        assert any(
            msg in output
            for msg in ["not compressed", "wrong extension", ".Z", "error"]
        )
    else:
        # May attempt to decompress anyway
        assert result == 0


def test_execute_empty_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with empty file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/dev/null"])

    # Should either succeed with empty output or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle empty file
        assert len(output.strip()) >= 0
    else:
        # Should fail if empty file not valid compressed format
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["not compressed", "invalid format", "error"]
        )


def test_execute_compressed_file_z_extension(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with .Z file (may not exist, but test the handling)
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=[tmp_name])

    # Should either decompress or fail with file not found
    if result == 0:
        output = capture.get()
        # Should show decompression result
        assert len(output.strip()) >= 0
    else:
        # Should fail if file doesn't exist or isn't compressed
        assert result == 1
        output = capture.get()
        assert any(
            msg in output
            for msg in ["No such file", "not found", "not compressed", "error"]
        )


def test_execute_force_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test -f option to force decompression
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-f", "/etc/hosts"])

    # Should either force decompression or fail gracefully
    if result == 0:
        output = capture.get()
        # Should attempt forced decompression
        assert len(output.strip()) >= 0
    else:
        # Should fail if forced decompression not possible
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["not compressed", "invalid format", "error"]
        )


def test_execute_verbose_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test -v option for verbose output
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-v", tmp_name])

    # Should either show verbose output or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show verbose decompression information
        assert (
            any(
                info in output
                for info in ["decompressing", "extracting", "uncompress", "done"]
            )
            or len(output.strip()) >= 0
        )
    else:
        # Should fail with verbose error information
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["not found", "not compressed", "error"])


def test_execute_keep_original_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test -k option to keep original file
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-k", tmp_name])

    # Should either keep original during decompression or fail gracefully
    if result == 0:
        output = capture.get()
        # Should decompress while keeping original
        assert len(output.strip()) >= 0
    else:
        # Should fail if file doesn't exist or option not supported
        assert result == 1


def test_execute_test_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test -t option to test compressed file integrity
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-t", tmp_name])

    # Should either test file integrity or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show test results
        assert (
            any(msg in output for msg in ["OK", "valid", "test", "integrity"])
            or len(output.strip()) >= 0
        )
    else:
        # Should fail if file doesn't exist or test fails
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["not found", "corrupt", "invalid", "error"]
        )


def test_execute_multiple_files(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with multiple files
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix="1.Z", delete=False) as tmp_file1:
            tmp_name1 = tmp_file1.name
        with tempfile.NamedTemporaryFile(suffix="2.Z", delete=False) as tmp_file2:
            tmp_name2 = tmp_file2.name
        result = command.execute(client=client, args=[tmp_name1, tmp_name2])

    # Should either decompress multiple files or fail gracefully
    if result == 0:
        output = capture.get()
        # Should process multiple files
        assert len(output.strip()) >= 0
    else:
        # Should fail if files don't exist
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["not found", "error"])


def test_execute_directory_argument(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with directory argument
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc"])

    # Should fail with directory error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output
        for msg in [
            "is a directory",
            "cannot decompress directory",
            "not a file",
            "error",
        ]
    )


def test_execute_permission_denied_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with permission denied file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/shadow"])

    # Should fail with permission error
    if result == 1:
        output = capture.get()
        assert any(
            msg in output
            for msg in ["permission denied", "cannot open", "access denied", "error"]
        )
    else:
        # May succeed if file is readable
        assert result == 0


def test_execute_output_to_stdout(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test -c option to output to stdout
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-c", tmp_name])

    # Should either output to stdout or fail gracefully
    if result == 0:
        output = capture.get()
        # Should output decompressed content to stdout
        assert len(output.strip()) >= 0
    else:
        # Should fail if file doesn't exist or isn't compressed
        assert result == 1


def test_execute_quiet_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test -q option for quiet mode
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-q", tmp_name])

    # Should either decompress quietly or fail gracefully
    if result == 0:
        output = capture.get()
        # Should have minimal output in quiet mode
        assert len(output.strip()) == 0 or len(output.strip()) < 50
    else:
        # Should fail quietly if file doesn't exist
        assert result == 1


def test_execute_recursive_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test -r option for recursive decompression
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-r", tempfile.mkdtemp()])

    # Should either decompress recursively or fail gracefully
    if result == 0:
        output = capture.get()
        # Should process directory recursively
        assert len(output.strip()) >= 0
    else:
        # Should fail if recursive option not supported or no compressed files found
        assert result == 1


def test_execute_invalid_compressed_format(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with file that has .Z extension but isn't actually compressed
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should fail with invalid format error
    if result == 1:
        output = capture.get()
        assert any(
            msg in output
            for msg in ["not compressed", "invalid format", "wrong format", "error"]
        )
    else:
        # May attempt to process anyway
        assert result == 0


def test_execute_large_compressed_file_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test with large compressed file (simulated)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/usr/bin/bash.Z"])

    # Should either handle large file or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle large file efficiently
        assert len(output.strip()) >= 0
    else:
        # Should fail if file doesn't exist or isn't compressed
        assert result == 1


def test_execute_backup_file_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test backup file creation behavior
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-b", tmp_name])

    # Should either create backup or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle backup creation
        assert len(output.strip()) >= 0
    else:
        # Should fail if backup option not supported or file doesn't exist
        assert result == 1


def test_execute_compression_ratio_display(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test compression ratio display
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-v", tmp_name])

    # Should either show compression ratio or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show decompression statistics
        assert (
            any(stat in output for stat in ["ratio", "size", "%", "bytes"])
            or len(output.strip()) >= 0
        )
    else:
        # Should fail if file doesn't exist
        assert result == 1


def test_execute_overwrite_protection(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test overwrite protection
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=[tmp_name])

    # Should either protect from overwrite or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle overwrite protection
        assert len(output.strip()) >= 0
    else:
        # Should fail if file doesn't exist or overwrite conflict
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["exists", "overwrite", "not found", "error"]
        )


def test_execute_symlink_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test symbolic link handling
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix="_link.Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=[tmp_name])

    # Should either follow symlink or handle appropriately
    if result == 0:
        output = capture.get()
        # Should follow symlink and decompress target
        assert len(output.strip()) >= 0
    else:
        # Should fail if symlink doesn't exist or target isn't compressed
        assert result == 1


def test_execute_filename_preservation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test filename preservation during decompression
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(
            suffix="_document.txt.Z", delete=False
        ) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=[tmp_name])

    # Should either preserve original filename or fail gracefully
    if result == 0:
        output = capture.get()
        # Should preserve filename without .Z extension
        assert len(output.strip()) >= 0
    else:
        # Should fail if file doesn't exist
        assert result == 1


def test_execute_timestamp_preservation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test timestamp preservation
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-p", tmp_name])

    # Should either preserve timestamps or fail gracefully
    if result == 0:
        output = capture.get()
        # Should preserve file timestamps
        assert len(output.strip()) >= 0
    else:
        # Should fail if option not supported or file doesn't exist
        assert result == 1


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/nonexistent.Z"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(msg in output for msg in ["not found", "error", "No such file"])


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=[tmp_name])

    # Should be memory efficient
    if result == 0:
        output = capture.get()
        # Should not consume excessive memory
        assert len(output) < 100000  # Reasonable output size limit
    else:
        assert result == 1


def test_execute_concurrent_execution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test concurrent execution safety
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=[tmp_name])

    # Should handle concurrent execution safely
    if result == 0:
        output = capture.get()
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_progress_indication(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test progress indication for large files
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix="_large.Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["-v", tmp_name])

    # Should either show progress or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show progress information
        assert (
            any(
                progress in output
                for progress in ["progress", "%", "decompressing", "done"]
            )
            or len(output.strip()) >= 0
        )
    else:
        # Should fail if file doesn't exist
        assert result == 1


def test_execute_format_detection(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test automatic format detection
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix="_test", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=[tmp_name])

    # Should either detect format automatically or fail gracefully
    if result == 0:
        output = capture.get()
        # Should detect and decompress format
        assert len(output.strip()) >= 0
    else:
        # Should fail if format not detected or file doesn't exist
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["format", "not found", "not compressed", "error"]
        )


def test_execute_invalid_option_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UncompressCommand,
):
    # Test handling of invalid options
    with command.shell.console.capture() as capture:
        with tempfile.NamedTemporaryFile(suffix=".Z", delete=False) as tmp_file:
            tmp_name = tmp_file.name
        result = command.execute(client=client, args=["--invalid-option", tmp_name])

    # Should handle invalid options appropriately
    if result == 1:
        output = capture.get()
        # Should show error for invalid option
        assert any(msg in output for msg in ["invalid", "unknown", "option"])
    else:
        # May ignore unknown options and proceed
        assert result == 0



"""Implementation of TimeoutCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag
from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class TimeoutCommand(Command):
    """Command for running other commands with a time limit."""

    name = "timeout"
    help = "Run a command with a time limit. Usage: timeout SECONDS COMMAND [ARGS...]"
    category = "Built-in Commands"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute the timeout command to run commands with time constraints."""
        if handle_help_flag(self, args):
            return 0
        if not args or len(args) < 3:
            self.console.print(
                "Usage: timeout [--preserve-status] SECONDS COMMAND [ARGS...]"
            )
            return 1
        try:
            seconds = float(args[0])
            if seconds <= 0:
                self.console.print("timeout: time must be positive")
                return 1
        except ValueError:
            self.console.print(f"timeout: invalid time interval '{args[0]}'")
            return 1
        cmd = args[1:]
        if args[0] == "--preserve-status":
            preserve_status = True
            cmd = args[2:]
        else:
            preserve_status = False
        process = client.exec(cmd, timeout=seconds)

        try:
            stdout, stderr = process.wait_output()
            if stdout:
                self.console.print(stdout, end="")
            if stderr:
                self.console.print(stderr, end="")
            return 0
        except ops.pebble.ChangeError:
            # This is expected when timeout occurs - return standard timeout exit code
            return 124
        except ops.pebble.TimeoutError:
            return 124
        except ops.pebble.ExecError as e:
            if e.stdout:
                self.console.print(e.stdout, end="")
            if e.stderr:
                self.console.print(e.stderr, end="")
            if preserve_status:
                return e.exit_code
            return 125



"""Implementation of EnvCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops
from rich.panel import Panel
from rich.table import Table

from ...utils.command_helpers import handle_help_flag
from ...utils.proc_reader import ProcReadError, read_proc_environ
from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class EnvCommand(Command):
    """Command for displaying environment variables from process files."""

    name = "env"
    help = "Show environment variables"
    category = "Built-in Commands"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute the env command to display environment variables."""
        if handle_help_flag(self, args):
            return 0
        console = self.console
        pids = ["self", "1"]  # Try self first, then init

        for pid in pids:
            try:
                env_vars = read_proc_environ(client, pid)

                table = Table(
                    show_header=True,
                    header_style="bold magenta",
                    box=None,
                    expand=False,
                )
                table.add_column("Variable", style="cyan", no_wrap=True)
                table.add_column("Value", style="green")

                for name, value in env_vars.items():
                    table.add_row(name, value)

                if table.row_count > 0:
                    console.print(
                        Panel(
                            table,
                            title=f"[b]Environment Variables from /proc/{pid}/environ[/b]",
                            style="cyan",
                        )
                    )
                    return 0
            except ProcReadError:  # noqa: PERF203
                # Continue to next PID if this one fails
                continue

        console.print(
            Panel(
                "Could not read environment variables from any source",
                title="[b red]env Error[/b red]",
                style="red",
            )
        )
        return 1



"""Integration tests for the AdduserCommand."""

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a AdduserCommand instance."""
    yield pebble_shell.commands.AdduserCommand(shell=shell)


def test_name(command: pebble_shell.commands.AdduserCommand):
    assert command.name == "adduser"


def test_category(command: pebble_shell.commands.AdduserCommand):
    assert command.category == "User Management"


def test_help(command: pebble_shell.commands.AdduserCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "adduser" in output
    assert any(
        phrase in output.lower() for phrase in ["add", "user", "create", "account"]
    )


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "adduser" in output
    assert any(
        phrase in output.lower() for phrase in ["add", "user", "create", "account"]
    )


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # adduser with no args should show usage
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should fail with usage message
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["usage", "Usage", "adduser", "username"])


def test_execute_add_user_basic(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test basic user addition
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["testuser"])

    # Should either succeed or fail due to permissions
    if result == 0:
        output = capture.get()
        # Should indicate user creation success
        assert len(output) >= 0
        if len(output) > 0:
            assert any(word in output.lower() for word in ["add", "creat", "success"])
    else:
        # Should fail if not root or user exists
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["permission", "denied", "root", "exist", "error"]
        )


def test_execute_system_user_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test system user creation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--system", "sysuser"])

    # Should either succeed creating system user or fail
    if result == 0:
        output = capture.get()
        # Should create system user
        assert len(output) >= 0
    else:
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["permission", "denied", "error", "exist"])


def test_execute_no_create_home_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation without home directory
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--no-create-home", "nohomeuser"])

    # Should either succeed without creating home or fail
    if result == 0:
        output = capture.get()
        # Should create user without home directory
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_disabled_password_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation with disabled password
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--disabled-password", "disableduser"]
        )

    # Should either succeed with disabled password or fail
    if result == 0:
        output = capture.get()
        # Should create user with disabled password
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_disabled_login_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation with disabled login
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--disabled-login", "nologinuser"]
        )

    # Should either succeed with disabled login or fail
    if result == 0:
        output = capture.get()
        # Should create user with disabled login
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_home_directory_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation with custom home directory
    custom_home = tempfile.mkdtemp()
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--home", custom_home, "customuser"]
        )

    # Should either succeed with custom home or fail
    if result == 0:
        output = capture.get()
        # Should create user with custom home
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_shell_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation with custom shell
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--shell", "/bin/bash", "bashuser"]
        )

    # Should either succeed with custom shell or fail
    if result == 0:
        output = capture.get()
        # Should create user with custom shell
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_uid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation with specific UID
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--uid", "1500", "uiduser"])

    # Should either succeed with specific UID or fail
    if result == 0:
        output = capture.get()
        # Should create user with specific UID
        assert len(output) >= 0
    else:
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["permission", "uid", "exist", "error"])


def test_execute_gid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation with specific GID
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--gid", "1500", "giduser"])

    # Should either succeed with specific GID or fail
    if result == 0:
        output = capture.get()
        # Should create user with specific GID
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_ingroup_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation in specific group
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--ingroup", "users", "groupuser"]
        )

    # Should either succeed adding to group or fail
    if result == 0:
        output = capture.get()
        # Should create user in specified group
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_gecos_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test user creation with GECOS field
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--gecos", "Test User,,,", "gecosuser"]
        )

    # Should either succeed with GECOS or fail
    if result == 0:
        output = capture.get()
        # Should create user with GECOS information
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_quiet_mode(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test quiet mode operation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--quiet", "quietuser"])

    # Should either succeed quietly or fail
    if result == 0:
        output = capture.get()
        # Should produce minimal output in quiet mode
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_force_badname_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test forcing bad username
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--force-badname", "bad-name"])

    # Should either succeed with forced bad name or fail
    if result == 0:
        output = capture.get()
        # Should create user with non-standard name
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_existing_user(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test adding existing user (should fail)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["root"])

    # Should fail with user exists error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["exist", "already", "error"])


def test_execute_invalid_username(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test invalid username
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["invalid@user"])

    # Should fail with invalid username error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "name", "error"])


def test_execute_invalid_uid(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test invalid UID
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--uid", "abc", "testuser"])

    # Should fail with invalid UID error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "uid", "number", "error"])


def test_execute_invalid_gid(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test invalid GID
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--gid", "xyz", "testuser"])

    # Should fail with invalid GID error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "gid", "number", "error"])


def test_execute_nonexistent_group(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test adding user to nonexistent group
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--ingroup", "nonexistentgroup", "testuser"]
        )

    # Should fail with group not found error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["group", "not found", "exist", "error"])


def test_execute_nonexistent_shell(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test setting nonexistent shell
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--shell", "/nonexistent/shell", "testuser"]
        )

    # Should either succeed or fail depending on validation
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["shell", "not found", "invalid", "error"])


def test_execute_invalid_home_directory(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test invalid home directory path
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["--home", "relative/path", "testuser"]
        )

    # Should either succeed or fail depending on validation
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_permission_denied(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test without sufficient permissions (non-root)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["permissionuser"])

    # Should either succeed or fail with permission error
    if result == 0:
        output = capture.get()
        # If running as root, should succeed
        assert len(output) >= 0
    else:
        # Should fail if not root
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["permission", "denied", "root", "error"])


def test_execute_invalid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test invalid option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--invalid-option", "testuser"])

    # Should fail with invalid option error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "unknown", "option", "usage"])


def test_execute_conflicting_options(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test conflicting options
    test_home = tempfile.mkdtemp()
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client,
            args=["--system", "--no-create-home", "--home", test_home, "testuser"],
        )

    # Should either handle conflicts or fail appropriately
    if result == 0:
        output = capture.get()
        # May resolve conflicts automatically
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["memoryuser"])

    # Should be memory efficient
    if result == 0:
        output = capture.get()
        # Should not consume excessive memory
        assert len(output) < 10000  # Reasonable output size limit
    else:
        assert result == 1


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--invalid", "testuser"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(msg in output for msg in ["invalid", "unknown", "error", "usage"])


def test_execute_signal_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test signal handling during user creation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["signaluser"])

    # Should handle signals appropriately
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_system_integration(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test system integration
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["integrationuser"])

    # Should integrate with system user management properly
    if result == 0:
        output = capture.get()
        # Should integrate with system
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_concurrent_execution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test concurrent execution safety
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["concurrentuser"])

    # Should handle concurrent execution safely
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_validation_comprehensive(
    client: ops.pebble.Client,
    command: pebble_shell.commands.AdduserCommand,
):
    # Test comprehensive input validation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["validuser"])

    # Should validate inputs comprehensively
    if result == 0:
        output = capture.get()
        # Should pass validation
        assert len(output) >= 0
    else:
        assert result == 1



"""Implementation of AdduserCommand."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags, safe_read_file
from .._base import Command
from .addgroup import AddgroupCommand

if TYPE_CHECKING:
    import shimmer


# Standard Unix shadow file values
LOCKED_ACCOUNT_HASH = "!"  # Standard indicator for locked/disabled account


# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class AdduserCommand(Command):
    """Implementation of adduser command."""

    name = "adduser"
    help = "Add a new user to the system"
    category = "User Management"

    def show_help(self):
        """Show command help."""
        help_text = """Add a new user to the system.

Usage: adduser [OPTIONS] USERNAME

Description:
    Add a new user to the system by modifying /etc/passwd, /etc/shadow,
    and optionally creating a home directory and default group.

Options:
    --uid UID           Specify user ID (default: auto-assign)
    --gid GID           Specify primary group ID
    --home DIR          Home directory (default: /home/USERNAME)
    --shell SHELL       Login shell (default: /bin/sh)
    --gecos GECOS       User information (full name, etc.)
    --system            Create a system user (UID < 1000)
    --no-create-home    Don't create home directory
    --disabled-password Create account with disabled password
    --quiet             Suppress informational messages
    -h, --help          Show this help message

Examples:
    adduser john                    # Create regular user
    adduser --uid 1001 jane         # Create user with specific UID
    adduser --system daemon         # Create system user
    adduser --shell /bin/bash bob   # Create user with bash shell
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the adduser command."""
        if handle_help_flag(self, args):
            return 0

        if not args:
            self.console.print("[red]adduser: missing username[/red]")
            return 1

        parse_result = parse_flags(
            args,
            {
                "uid": int,  # user ID
                "gid": int,  # group ID
                "home": str,  # home directory
                "shell": str,  # shell
                "gecos": str,  # user info
                "system": bool,  # system user
                "no-create-home": bool,  # don't create home
                "disabled-password": bool,  # disabled password
                "quiet": bool,  # quiet mode
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        if not positional_args:
            self.console.print("[red]adduser: missing username[/red]")
            return 1

        username = positional_args[0]
        uid = flags.get("uid")
        gid = flags.get("gid")
        home_dir = flags.get("home", f"/home/{username}")
        shell = flags.get("shell", "/bin/sh")
        gecos = flags.get("gecos", "")
        system_user = flags.get("system", False)
        no_create_home = flags.get("no-create-home", False)
        disabled_password = flags.get("disabled-password", False)
        quiet = flags.get("quiet", False)

        try:
            return self._create_user(
                client,
                username,
                uid,
                gid,
                home_dir,
                shell,
                gecos,
                system_user,
                no_create_home,
                disabled_password,
                quiet,
            )

        except Exception as e:
            self.console.print(f"[red]adduser: {e}[/red]")
            return 1

    def _create_user(
        self,
        client: ClientType,
        username: str,
        uid: int | None,
        gid: int | None,
        home_dir: str,
        shell: str,
        gecos: str,
        system_user: bool,
        no_create_home: bool,
        disabled_password: bool,
        quiet: bool,
    ) -> int:
        """Create a new user."""
        # Read existing /etc/passwd
        passwd_content = safe_read_file(client, "/etc/passwd")
        if passwd_content is None:
            self.console.print("[red]adduser: cannot read /etc/passwd[/red]")
            return 1

        # Check if user already exists
        for line in passwd_content.strip().split("\n"):
            if line and line.split(":")[0] == username:
                self.console.print(
                    f"[red]adduser: user '{username}' already exists[/red]"
                )
                return 1

        # Determine UID
        if uid is None:
            uid = self._find_available_uid(passwd_content, system_user)

        # Check if UID is already in use
        for line in passwd_content.strip().split("\n"):
            if line and len(line.split(":")) >= 3:
                existing_uid = line.split(":")[2]
                if existing_uid == str(uid):
                    self.console.print(f"[red]adduser: UID {uid} already in use[/red]")
                    return 1

        # Determine GID - create group if not specified
        if gid is None:
            # Create a group with the same name as the user
            group_result = self._create_user_group(client, username, system_user, quiet)
            if group_result != 0:
                return group_result
            gid = uid  # Use same ID for group

        # Validate that GID exists
        group_content = safe_read_file(client, "/etc/group")
        if group_content is not None:
            gid_exists = False
            for line in group_content.strip().split("\n"):
                if (
                    line
                    and len(line.split(":")) >= 3
                    and line.split(":")[2] == str(gid)
                ):
                    gid_exists = True
                    break

            if not gid_exists:
                self.console.print(
                    f"[red]adduser: group with GID {gid} does not exist[/red]"
                )
                return 1

        # Create passwd entry
        passwd_entry = f"{username}:x:{uid}:{gid}:{gecos}:{home_dir}:{shell}"

        # Read existing /etc/shadow
        shadow_content = safe_read_file(client, "/etc/shadow")
        if shadow_content is None:
            # Create minimal shadow if it doesn't exist
            shadow_content = ""

        # Create shadow entry
        if disabled_password:
            password_hash = LOCKED_ACCOUNT_HASH
        else:
            # Generate a random password hash (user will need to set password)
            password_hash = LOCKED_ACCOUNT_HASH  # Account locked until password is set

        # Get days since epoch for password age
        days_since_epoch = int(time.time() // 86400)

        shadow_entry = f"{username}:{password_hash}:{days_since_epoch}:0:99999:7:::"

        # Show what would be done
        if not quiet:
            self.console.print(
                f"[green]Adding user '{username}' (UID {uid}, GID {gid})[/green]"
            )

        self.console.print("[yellow]adduser: would write to /etc/passwd:[/yellow]")
        self.console.print(f"[dim]{passwd_entry}[/dim]")

        self.console.print("[yellow]adduser: would write to /etc/shadow:[/yellow]")
        self.console.print(f"[dim]{shadow_entry}[/dim]")

        # Create home directory
        if not no_create_home and not system_user:
            self.console.print(
                f"[yellow]adduser: would create home directory: {home_dir}[/yellow]"
            )
            self.console.print(
                f"[yellow]adduser: would set ownership to {uid}:{gid}[/yellow]"
            )

        if not quiet:
            self.console.print(f"[green]User '{username}' created successfully[/green]")
            if disabled_password:
                self.console.print(
                    "[yellow]Account created with disabled password[/yellow]"
                )
            else:
                self.console.print(
                    "[yellow]User account locked - use passwd to set password[/yellow]"
                )

        return 0

    def _create_user_group(
        self, client: ClientType, username: str, system_user: bool, quiet: bool
    ) -> int:
        """Create a group for the user."""
        # Use the addgroup command to create the group
        addgroup_cmd = AddgroupCommand(self.shell)

        args = ["--quiet"] if quiet else []
        if system_user:
            args.append("--system")
        args.append(username)

        return addgroup_cmd.execute(client, args)

    def _find_available_uid(self, passwd_content: str, system_user: bool) -> int:
        """Find next available UID."""
        used_uids = set()

        for line in passwd_content.strip().split("\n"):
            if line and len(line.split(":")) >= 3:
                try:
                    uid = int(line.split(":")[2])
                    used_uids.add(uid)
                except ValueError:
                    continue

        # System users: 100-999, regular users: 1000+
        start_uid = 100 if system_user else 1000
        max_uid = 999 if system_user else 65534

        for uid in range(start_uid, max_uid + 1):
            if uid not in used_uids:
                return uid

        raise Exception(f"No available UID in range {start_uid}-{max_uid}")



"""Integration tests for the PushCommand."""

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a PushCommand instance."""
    yield pebble_shell.commands.PushCommand(shell=shell)


def test_name(command: pebble_shell.commands.PushCommand):
    assert command.name == "pebble-push"


def test_category(command: pebble_shell.commands.PushCommand):
    assert command.category == "Pebble Management"


def test_help(command: pebble_shell.commands.PushCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "push" in output
    assert "Transfer files to container" in output
    assert "--create-dirs" in output
    assert "--permissions" in output


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "push" in output
    assert "Transfer files to container" in output


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # push with no args should fail and show usage
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should fail with usage message
    assert result == 1
    output = capture.get()
    assert "push <source-path> <dest-path>" in output


def test_execute_missing_dest_path(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # push with only source path should fail
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/local/file.txt"])

    # Should fail with missing destination error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output
        for msg in ["missing destination", "push <source-path> <dest-path>", "usage"]
    )


def test_execute_simple_file_push(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test pushing simple file to container
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/passwd"]
        )

    # Should either succeed with file transfer or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show transfer completion message
        assert any(
            msg in output
            for msg in ["File pushed successfully", "Transfer completed", "uploaded"]
        )
    else:
        # Should fail if file doesn't exist or Pebble unavailable
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["Push operation failed", "error", "failed"]
        )


def test_execute_nonexistent_source_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test with nonexistent source file
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/nonexistent/file.txt", "/container/file.txt"]
        )

    # Should fail with file not found error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output
        for msg in ["Push operation failed", "file not found", "No such file", "error"]
    )


def test_execute_directory_push(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test pushing directory to container
    temp_dir = tempfile.mkdtemp()
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[temp_dir, "/container/tmp"])

    # Should either succeed with directory transfer or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show directory transfer completion
        assert any(
            msg in output
            for msg in [
                "Directory pushed successfully",
                "Transfer completed",
                "uploaded",
            ]
        )
    else:
        assert result == 1


def test_execute_create_dirs_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test --create-dirs option
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client,
            args=["--create-dirs", "/etc/passwd", "/container/deep/dir/passwd"],
        )

    # Should either succeed creating directories or fail gracefully
    if result == 0:
        output = capture.get()
        # Should create directories and transfer file
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_permissions_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test --permissions option
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client,
            args=["--permissions", "755", "/etc/passwd", "/container/passwd"],
        )

    # Should either succeed with permission setting or fail gracefully
    if result == 0:
        output = capture.get()
        # Should set permissions and transfer file
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_invalid_permissions(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test with invalid permissions
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client,
            args=["--permissions", "999", "/etc/passwd", "/container/passwd"],
        )

    # Should fail with invalid permissions error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output
        for msg in ["invalid permissions", "Push operation failed", "error"]
    )


def test_execute_overwrite_existing_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test overwriting existing destination file
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/existing-file"]
        )

    # Should either succeed overwriting file or fail gracefully
    if result == 0:
        output = capture.get()
        # Should complete transfer with overwrite
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_permission_denied_source(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test with permission denied on source file
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/root/.ssh/id_rsa", "/container/key"]
        )

    # Should fail with permission error
    if result == 1:
        output = capture.get()
        assert any(
            msg in output
            for msg in ["permission denied", "Push operation failed", "error"]
        )
    else:
        # May succeed if file is readable
        assert result == 0


def test_execute_permission_denied_destination(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test with permission denied on destination
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/root/restricted"]
        )

    # Should fail with permission error
    if result == 1:
        output = capture.get()
        assert any(
            msg in output
            for msg in ["permission denied", "Push operation failed", "error"]
        )
    else:
        # May succeed if destination is writable
        assert result == 0


def test_execute_large_file_transfer(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test transferring large file
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/usr/bin/bash", "/container/bash"]
        )

    # Should either succeed with large file transfer or fail gracefully
    if result == 0:
        output = capture.get()
        # Should complete large file transfer
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_binary_file_transfer(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test transferring binary file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls", "/container/ls"])

    # Should handle binary file transfer
    if result == 0:
        output = capture.get()
        # Should complete binary transfer
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_text_file_transfer(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test transferring text file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts", "/container/hosts"])

    # Should handle text file transfer
    if result == 0:
        output = capture.get()
        # Should complete text transfer
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_empty_file_transfer(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test transferring empty file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/dev/null", "/container/empty"])

    # Should handle empty file transfer
    if result == 0:
        output = capture.get()
        # Should complete empty file transfer
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_symlink_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test handling of symbolic links
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin", "/container/bin-link"])

    # Should handle symlink appropriately
    if result == 0:
        output = capture.get()
        # Should transfer symlink or target
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_absolute_paths(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test with absolute paths
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/passwd-copy"]
        )

    # Should handle absolute paths
    if result == 0:
        output = capture.get()
        # Should complete transfer with absolute paths
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_relative_paths(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test with relative paths
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["../etc/passwd", "passwd-rel"])

    # Should handle relative paths
    if result == 0:
        output = capture.get()
        # Should resolve and transfer relative paths
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_special_characters_in_paths(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test paths with special characters
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/file with spaces & chars"]
        )

    # Should handle special characters in paths
    if result == 0:
        output = capture.get()
        # Should handle special characters
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_nested_directory_push(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test pushing nested directory structure
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/usr/share/doc", "/container/doc"]
        )

    # Should handle nested directory transfer
    if result == 0:
        output = capture.get()
        # Should transfer entire directory tree
        assert any(
            msg in output
            for msg in ["Directory pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_file_metadata_preservation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test file metadata preservation during transfer
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/passwd-meta"]
        )

    # Should preserve file metadata
    if result == 0:
        output = capture.get()
        # Should maintain file attributes
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_concurrent_push_operations(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test concurrent push operations
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/hosts", "/container/concurrent-hosts"]
        )

    # Should handle concurrent operations
    if result == 0:
        output = capture.get()
        # Should complete concurrent transfer
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_network_interruption_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test handling of network interruptions
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/network-test"]
        )

    # Should handle network issues gracefully
    if result == 1:
        output = capture.get()
        assert any(
            msg in output for msg in ["Push operation failed", "network", "error"]
        )
    else:
        assert result == 0


def test_execute_disk_space_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test handling of insufficient disk space
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/usr/bin/bash", "/container/space-test"]
        )

    # Should handle disk space issues
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["Push operation failed", "space", "error"])
    else:
        assert result == 0


def test_execute_api_error_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test API error handling when Pebble is unavailable
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/api-test"]
        )

    # Should handle API errors gracefully
    if result == 1:
        output = capture.get()
        assert any(
            msg in output for msg in ["Push operation failed", "error", "failed"]
        )
    else:
        # Should succeed if Pebble API is available
        assert result == 0


def test_execute_progress_reporting(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test progress reporting for large transfers
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/usr/bin/bash", "/container/progress-test"]
        )

    # Should report transfer progress
    if result == 0:
        output = capture.get()
        # Should show progress information
        assert any(
            msg in output
            for msg in ["File pushed successfully", "Transfer completed", "bytes"]
        )
    else:
        assert result == 1


def test_execute_checksum_verification(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test checksum verification of transferred files
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/checksum-test"]
        )

    # Should verify file integrity
    if result == 0:
        output = capture.get()
        # Should complete with integrity check
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_cleanup_on_failure(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test cleanup on transfer failure
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/nonexistent/file", "/container/cleanup-test"]
        )

    # Should clean up on failure
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["Push operation failed", "error"])


def test_execute_recursive_directory_push(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test recursive directory pushing
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/systemd", "/container/systemd"]
        )

    # Should handle recursive directory transfer
    if result == 0:
        output = capture.get()
        # Should transfer entire directory recursively
        assert any(
            msg in output
            for msg in ["Directory pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_transfer_cancellation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test transfer cancellation handling
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/cancel-test"]
        )

    # Should handle cancellation appropriately
    if result == 0:
        output = capture.get()
        # Should complete or handle cancellation
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_owner_group_setting(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test setting owner and group during push
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client,
            args=["--permissions", "644", "/etc/passwd", "/container/owned-file"],
        )

    # Should set ownership appropriately
    if result == 0:
        output = capture.get()
        # Should complete with ownership setting
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_atomic_transfer(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test atomic transfer operations
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/atomic-test"]
        )

    # Should perform atomic transfer
    if result == 0:
        output = capture.get()
        # Should complete atomically
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1


def test_execute_backup_existing_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.PushCommand,
):
    # Test backing up existing files before overwrite
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", "/container/backup-test"]
        )

    # Should handle existing file appropriately
    if result == 0:
        output = capture.get()
        # Should complete with backup handling
        assert any(
            msg in output for msg in ["File pushed successfully", "Transfer completed"]
        )
    else:
        assert result == 1



"""Tests for shell parser functionality."""

from unittest.mock import patch

import pytest

from pebble_shell.utils.parser import (
    CommandType,
    ShellParser,
    ShellVariables,
    get_shell_parser,
    init_shell_parser,
)


class TestShellVariables:
    """Test cases for ShellVariables class."""

    @pytest.fixture
    def variables(self):
        """Create ShellVariables instance."""
        return ShellVariables()

    def test_init_default_variables(self, variables):
        """Test initialization with default variables."""
        assert variables.get_variable("PWD") == "/"
        assert variables.get_variable("USER") == "root"
        assert variables.get_variable("HOME") == "/root"
        assert variables.get_variable("?") == "0"

    def test_set_and_get_variable(self, variables):
        """Test setting and getting variables."""
        variables.set_variable("MYVAR", "myvalue")
        assert variables.get_variable("MYVAR") == "myvalue"

    def test_get_nonexistent_variable(self, variables):
        """Test getting non-existent variable."""
        assert variables.get_variable("NONEXISTENT") == ""

    @patch("os.environ.get")
    def test_get_environment_variable(self, mock_env_get, variables):
        """Test getting environment variable."""
        mock_env_get.return_value = "env_value"
        assert variables.get_variable("ENV_VAR") == "env_value"
        mock_env_get.assert_called_once_with("ENV_VAR", "")

    def test_expand_variables_simple(self, variables):
        """Test simple variable expansion."""
        variables.set_variable("USER", "testuser")
        result = variables.expand_variables("Hello $USER")
        assert result == "Hello testuser"

    def test_expand_variables_braced(self, variables):
        """Test braced variable expansion."""
        variables.set_variable("USER", "testuser")
        result = variables.expand_variables("Hello ${USER}!")
        assert result == "Hello testuser!"

    def test_expand_variables_multiple(self, variables):
        """Test multiple variable expansion."""
        variables.set_variable("USER", "testuser")
        variables.set_variable("HOME", "/home/test")
        result = variables.expand_variables("$USER lives in $HOME")
        assert result == "testuser lives in /home/test"

    def test_expand_variables_no_expansion(self, variables):
        """Test text with no variables."""
        result = variables.expand_variables("No variables here")
        assert result == "No variables here"

    def test_expand_variables_nonexistent(self, variables):
        """Test expansion of non-existent variable."""
        result = variables.expand_variables("Hello $NONEXISTENT")
        assert result == "Hello "

    def test_update_pwd(self, variables):
        """Test updating PWD variable."""
        variables.update_pwd("/new/path")
        assert variables.get_variable("PWD") == "/new/path"

    def test_exit_code_variable(self, variables):
        """Test exit code special variable."""
        variables.last_exit_code = 42
        assert variables.get_variable("?") == "42"


class TestShellParser:
    """Test cases for ShellParser class."""

    @pytest.fixture
    def parser(self):
        """Create ShellParser instance."""
        return ShellParser()

    def test_init(self, parser):
        """Test parser initialization."""
        assert parser.variables is not None
        assert isinstance(parser.variables, ShellVariables)

    def test_parse_simple_command(self, parser):
        """Test parsing simple command."""
        commands = parser.parse_command_line("ls -la")

        assert len(commands) == 1
        assert commands[0].command == "ls"
        assert commands[0].args == ["-la"]
        assert commands[0].type == CommandType.SIMPLE

    def test_parse_empty_command(self, parser):
        """Test parsing empty command."""
        commands = parser.parse_command_line("")
        assert len(commands) == 0

    def test_parse_pipe_commands(self, parser):
        """Test parsing piped commands."""
        commands = parser.parse_command_line("ps | grep python")

        assert len(commands) == 2
        assert commands[0].command == "ps"
        assert commands[0].type == CommandType.PIPE
        assert commands[0].next_command == commands[1]
        assert commands[1].command == "grep"
        assert commands[1].args == ["python"]

    def test_parse_output_redirection(self, parser):
        """Test parsing output redirection."""
        commands = parser.parse_command_line("ls > output.txt")

        assert len(commands) == 1
        assert commands[0].command == "ls"
        assert commands[0].type == CommandType.REDIRECT_OUT
        assert commands[0].target == "output.txt"

    def test_parse_append_redirection(self, parser):
        """Test parsing append redirection."""
        commands = parser.parse_command_line("echo hello >> log.txt")

        assert len(commands) == 1
        assert commands[0].command == "echo"
        assert commands[0].args == ["hello"]
        assert commands[0].type == CommandType.REDIRECT_APPEND
        assert commands[0].target == "log.txt"

    def test_parse_input_redirection(self, parser):
        """Test parsing input redirection."""
        commands = parser.parse_command_line("wc < input.txt")

        assert len(commands) == 1
        assert commands[0].command == "wc"
        assert commands[0].type == CommandType.REDIRECT_IN
        assert commands[0].target == "input.txt"

    def test_parse_and_operator(self, parser):
        """Test parsing AND operator."""
        commands = parser.parse_command_line("mkdir test && cd test")

        assert len(commands) == 2
        assert commands[0].command == "mkdir"
        assert commands[0].args == ["test"]
        assert commands[0].type == CommandType.AND
        assert commands[1].command == "cd"
        assert commands[1].args == ["test"]

    def test_parse_or_operator(self, parser):
        """Test parsing OR operator."""
        commands = parser.parse_command_line("ls /nonexistent || echo failed")

        assert len(commands) == 2
        assert commands[0].command == "ls"
        assert commands[0].args == ["/nonexistent"]
        assert commands[0].type == CommandType.OR
        assert commands[1].command == "echo"
        assert commands[1].args == ["failed"]

    def test_parse_semicolon_separator(self, parser):
        """Test parsing semicolon separator."""
        commands = parser.parse_command_line("ls; pwd; whoami")

        assert len(commands) == 3
        assert commands[0].command == "ls"
        assert commands[0].type == CommandType.SEMICOLON
        assert commands[1].command == "pwd"
        assert commands[1].type == CommandType.SEMICOLON
        assert commands[2].command == "whoami"

    def test_parse_complex_pipeline(self, parser):
        """Test parsing complex pipeline with redirection."""
        commands = parser.parse_command_line("ps | grep python | wc -l > count.txt")

        assert len(commands) == 3
        assert commands[0].command == "ps"
        assert commands[0].type == CommandType.PIPE
        assert commands[1].command == "grep"
        assert commands[1].type == CommandType.PIPE
        assert commands[2].command == "wc"
        assert commands[2].args == ["-l"]
        assert commands[2].type == CommandType.REDIRECT_OUT
        assert commands[2].target == "count.txt"

    def test_parse_variable_expansion(self, parser):
        """Test variable expansion in parsing."""
        parser.set_variable("USER", "testuser")
        commands = parser.parse_command_line("echo Hello $USER")

        assert len(commands) == 1
        assert commands[0].command == "echo"
        assert commands[0].args == ["Hello", "testuser"]

    @patch("glob.glob")
    def test_parse_glob_expansion(self, mock_glob, parser):
        """Test glob expansion in parsing."""
        mock_glob.return_value = ["file1.txt", "file2.txt"]
        commands = parser.parse_command_line("ls *.txt")

        assert len(commands) == 1
        assert commands[0].command == "ls"
        assert commands[0].args == ["file1.txt", "file2.txt"]

    @patch("glob.glob")
    def test_parse_glob_no_matches(self, mock_glob, parser):
        """Test glob with no matches."""
        mock_glob.return_value = []
        commands = parser.parse_command_line("ls *.nonexistent")

        assert len(commands) == 1
        assert commands[0].command == "ls"
        assert commands[0].args == ["*.nonexistent"]  # Keep original if no matches

    def test_set_and_get_variable(self, parser):
        """Test setting and getting variables."""
        parser.set_variable("TESTVAR", "testvalue")
        assert parser.get_variable("TESTVAR") == "testvalue"

    def test_update_pwd(self, parser):
        """Test updating PWD."""
        parser.update_pwd("/new/directory")
        assert parser.get_variable("PWD") == "/new/directory"

    def test_set_exit_code(self, parser):
        """Test setting exit code."""
        parser.set_exit_code(42)
        assert parser.get_variable("?") == "42"

    def test_parse_quoted_arguments(self, parser):
        """Test parsing commands with quoted arguments."""
        commands = parser.parse_command_line("echo \"hello world\" 'single quotes'")

        assert len(commands) == 1
        assert commands[0].command == "echo"
        assert commands[0].args == ["hello world", "single quotes"]

    def test_parse_pipe_vs_or(self, parser):
        """Test distinguishing between pipe and OR operators."""
        # Test pipe
        commands = parser.parse_command_line("ps | grep test")
        assert len(commands) == 2
        assert commands[0].type == CommandType.PIPE

        # Test OR
        commands = parser.parse_command_line("false || echo backup")
        assert len(commands) == 2
        assert commands[0].type == CommandType.OR


class TestParserModule:
    """Test module-level parser functions."""

    def test_get_shell_parser_singleton(self):
        """Test that get_shell_parser returns singleton."""
        parser1 = get_shell_parser()
        parser2 = get_shell_parser()

        assert parser1 is parser2

    def test_init_shell_parser(self):
        """Test initializing shell parser."""
        variables = ShellVariables()
        parser = init_shell_parser(variables)

        assert isinstance(parser, ShellParser)
        assert parser.variables is variables

        # Should return same instance
        same_parser = get_shell_parser()
        assert parser is same_parser



"""Integration tests for the ReformineCommand."""

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a ReformineCommand instance."""
    yield pebble_shell.commands.ReformineCommand(shell=shell)


def test_name(command: pebble_shell.commands.ReformineCommand):
    assert command.name == "reformime"


def test_category(command: pebble_shell.commands.ReformineCommand):
    assert command.category == "System"


def test_help(command: pebble_shell.commands.ReformineCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert any(
        phrase in output.lower()
        for phrase in ["reform", "mime", "message", "email", "reformat"]
    )


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert any(
        phrase in output.lower() for phrase in ["reform", "mime", "message", "usage"]
    )


def test_execute_no_arguments(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with no arguments
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should fail with missing arguments error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output for msg in ["argument", "required", "usage", "error", "option"]
    )


def test_execute_extract_mime_structure(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test extracting MIME structure
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s"])

    # Should either succeed extracting structure or fail with input error
    if result == 0:
        output = capture.get()
        # Should extract MIME structure
        assert len(output) >= 0
    else:
        # Should fail with input or structure error
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["input", "structure", "mime", "error"])


def test_execute_extract_headers(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test extracting MIME headers
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-i"])

    # Should either succeed extracting headers or fail
    if result == 0:
        output = capture.get()
        # Should extract headers
        assert len(output) >= 0
    else:
        # Should fail with input or header error
        assert result == 1


def test_execute_extract_specific_section(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test extracting specific MIME section
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-x", "1"])

    # Should either succeed extracting section or fail
    if result == 0:
        output = capture.get()
        # Should extract specified section
        assert len(output) >= 0
    else:
        # Should fail with input or section error
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["section", "not found", "mime", "error"])


def test_execute_decode_quoted_printable(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test decoding quoted-printable content
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-q"])

    # Should either succeed decoding or fail
    if result == 0:
        output = capture.get()
        # Should decode quoted-printable
        assert len(output) >= 0
    else:
        # Should fail with input or decoding error
        assert result == 1


def test_execute_decode_base64(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test decoding base64 content
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-6"])

    # Should either succeed decoding base64 or fail
    if result == 0:
        output = capture.get()
        # Should decode base64
        assert len(output) >= 0
    else:
        # Should fail with input or decoding error
        assert result == 1


def test_execute_extract_text_content(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test extracting text content
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t"])

    # Should either succeed extracting text or fail
    if result == 0:
        output = capture.get()
        # Should extract text content
        assert len(output) >= 0
    else:
        # Should fail with input or text error
        assert result == 1


def test_execute_rewrite_addresses(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test rewriting email addresses
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["-R", "old@example.com", "new@example.com"]
        )

    # Should either succeed rewriting addresses or fail
    if result == 0:
        output = capture.get()
        # Should rewrite addresses
        assert len(output) >= 0
    else:
        # Should fail with input or rewrite error
        assert result == 1


def test_execute_with_input_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with input file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/etc/hostname"])

    # Should either succeed with file input or fail
    if result == 0:
        output = capture.get()
        # Should process file input
        assert len(output) >= 0
    else:
        # Should fail with file not found or format error
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["not found", "error", "file", "mime"])


def test_execute_nonexistent_input_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with non-existent input file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/nonexistent/mime.msg"])

    # Should fail with file not found error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output for msg in ["not found", "no such file", "error", "cannot open"]
    )


def test_execute_directory_instead_of_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with directory instead of file
    temp_dir = tempfile.mkdtemp()
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", temp_dir])

    # Should fail with directory error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["directory", "not a file", "error", "invalid"])


def test_execute_permission_denied_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with permission denied file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/root/.ssh/id_rsa"])

    # Should either succeed or fail with permission error
    if result == 0:
        output = capture.get()
        # Should read file if permitted
        assert len(output) >= 0
    else:
        # Should fail with permission error
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["permission", "denied", "access", "error"])


def test_execute_invalid_mime_format(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with invalid MIME format
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["-s", "/etc/passwd"]
        )  # Not MIME format

    # Should either succeed or fail with format error
    if result == 0:
        output = capture.get()
        # Should handle non-MIME format
        assert len(output) >= 0
    else:
        # Should fail with format error
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["format", "invalid", "mime", "error"])


def test_execute_extract_nonexistent_section(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test extracting non-existent section
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-x", "999", "/etc/hostname"])

    # Should fail with section not found error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output for msg in ["section", "not found", "invalid", "error", "999"]
    )


def test_execute_invalid_section_number(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with invalid section number
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-x", "abc"])

    # Should fail with invalid section number error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "section", "number", "error"])


def test_execute_negative_section_number(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with negative section number
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-x", "-1"])

    # Should fail with invalid section number error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "section", "negative", "error"])


def test_execute_stdin_input(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test reading from stdin
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "-"])

    # Should either succeed reading stdin or fail
    if result == 0:
        output = capture.get()
        # Should read from stdin
        assert len(output) >= 0
    else:
        # Should fail with stdin or format error
        assert result == 1


def test_execute_multiple_operations(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test multiple operations together
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "-i", "/etc/hostname"])

    # Should either succeed with multiple operations or fail
    if result == 0:
        output = capture.get()
        # Should perform multiple operations
        assert len(output) >= 0
    else:
        # Should fail with operation conflict or file error
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["conflict", "multiple", "operation", "error"]
        )


def test_execute_decode_with_charset(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test decoding with charset specification
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-c", "utf-8", "-t"])

    # Should either succeed with charset or fail
    if result == 0:
        output = capture.get()
        # Should decode with specified charset
        assert len(output) >= 0
    else:
        # Should fail with charset or input error
        assert result == 1


def test_execute_invalid_charset(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with invalid charset
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-c", "invalid-charset", "-t"])

    # Should fail with invalid charset error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "charset", "encoding", "error"])


def test_execute_extract_attachments(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test extracting attachments
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a"])

    # Should either succeed extracting attachments or fail
    if result == 0:
        output = capture.get()
        # Should extract attachments
        assert len(output) >= 0
    else:
        # Should fail with input or attachment error
        assert result == 1


def test_execute_print_mime_structure(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test printing MIME structure
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-D"])

    # Should either succeed printing structure or fail
    if result == 0:
        output = capture.get()
        # Should print MIME structure debug info
        assert len(output) >= 0
    else:
        # Should fail with input or debug error
        assert result == 1


def test_execute_invalid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test invalid option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-z"])

    # Should fail with invalid option error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "unknown", "option", "usage"])


def test_execute_empty_mime_message(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with empty MIME message
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/dev/null"])

    # Should either succeed with empty message or fail
    if result == 0:
        output = capture.get()
        # Should handle empty message
        assert len(output) >= 0
    else:
        # Should fail with empty or format error
        assert result == 1


def test_execute_large_mime_message(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with large MIME message
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/dev/zero"])

    # Should either handle large messages or fail appropriately
    if result == 0:
        output = capture.get()
        # Should handle large messages efficiently
        assert len(output) >= 0
    else:
        # Should fail with size or format error
        assert result == 1


def test_execute_binary_content_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test with binary content
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/bin/ls"])

    # Should either handle binary content or fail
    if result == 0:
        output = capture.get()
        # Should handle binary content
        assert len(output) >= 0
    else:
        # Should fail with binary or format error
        assert result == 1


def test_execute_multipart_message_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test multipart MIME message handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s"])

    # Should either succeed with multipart or fail
    if result == 0:
        output = capture.get()
        # Should handle multipart messages
        assert len(output) >= 0
    else:
        # Should fail with input or multipart error
        assert result == 1


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-x", "abc"])  # Trigger error

    # Should be memory efficient even on errors
    assert result == 1
    output = capture.get()
    # Should not consume excessive memory
    assert len(output) < 10000  # Reasonable error message size


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/invalid/path"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(
        msg in output for msg in ["not found", "error", "invalid", "cannot open"]
    )


def test_execute_signal_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test signal handling during MIME processing
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/etc/hostname"])

    # Should handle signals appropriately
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        # Should fail with file not found or format error
        assert result == 1


def test_execute_output_formatting(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test output formatting
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/etc/hostname"])

    # Should either succeed with proper formatting or fail
    if result == 0:
        output = capture.get()
        # Should format output properly
        assert len(output) >= 0
        if len(output) > 0:
            # Should contain formatted MIME information
            lines = output.split("\n")
            assert len(lines) >= 1
    else:
        # Should fail with file not found or format error
        assert result == 1


def test_execute_encoding_detection(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test encoding detection
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "/etc/hostname"])

    # Should either succeed detecting encoding or fail
    if result == 0:
        output = capture.get()
        # Should detect and handle encoding
        assert len(output) >= 0
    else:
        # Should fail with detection or format error
        assert result == 1


def test_execute_cross_platform_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ReformineCommand,
):
    # Test cross-platform compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/etc/hostname"])

    # Should work across different platforms
    if result == 0:
        output = capture.get()
        # Should handle platform differences
        assert len(output) >= 0
    else:
        # Should fail consistently across platforms
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["not found", "error", "format", "mime"])



"""Implementation of iplink command (read-only)."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from ...utils.command_helpers import handle_help_flag, safe_read_file
from ...utils.proc_reader import ProcReadError, read_proc_file
from .._base import Command

if TYPE_CHECKING:
    import ops
    import shimmer


# TODO: Use the prototype from Shimmer.
ClientType = Union["ops.pebble.Client", "shimmer.PebbleCliClient"]


class IplinkCommand(Command):
    """Implementation of iplink command (read-only)."""

    name = "iplink"
    help = "Display network interfaces"
    category = "Network"

    def show_help(self):
        """Show command help."""
        help_text = """Display network interfaces.

Usage: iplink [OPTIONS]

Description:
    Display network interface information. Read-only version.
    Simplified version of 'ip link show'.

Options:
    -h, --help      Show this help message

Examples:
    iplink          # Show all network interfaces
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the iplink command."""
        if handle_help_flag(self, args):
            return 0

        try:
            # Read network interfaces from /proc/net/dev
            try:
                net_dev = read_proc_file(client, "/proc/net/dev")
            except ProcReadError:
                self.console.print(
                    "[red]iplink: cannot read network interface information[/red]"
                )
                return 1

            lines = net_dev.strip().split("\n")

            index = 1
            for line in lines[2:]:  # Skip header lines
                if ":" in line:
                    iface_name = line.split(":")[0].strip()
                    if iface_name:
                        # Get interface flags and info
                        try:
                            operstate_path = f"/sys/class/net/{iface_name}/operstate"
                            operstate = safe_read_file(
                                client, operstate_path, self.shell
                            )
                            state = (
                                "UP"
                                if operstate and operstate.strip() == "up"
                                else "DOWN"
                            )
                        except Exception:
                            # Broad exception needed for network interface probing
                            state = "UNKNOWN"

                        try:
                            mtu_path = f"/sys/class/net/{iface_name}/mtu"
                            mtu = safe_read_file(client, mtu_path, self.shell)
                            mtu_value = mtu.strip() if mtu else "1500"
                        except Exception:
                            # Broad exception needed for network interface probing
                            mtu_value = "1500"

                        try:
                            addr_path = f"/sys/class/net/{iface_name}/address"
                            mac_addr = safe_read_file(client, addr_path, self.shell)
                            mac_addr = (
                                mac_addr.strip() if mac_addr else "00:00:00:00:00:00"
                            )
                        except Exception:
                            # Broad exception needed for network interface probing
                            mac_addr = "00:00:00:00:00:00"

                        flags = "BROADCAST,MULTICAST"
                        if state == "UP":
                            flags += ",UP,LOWER_UP"
                        if iface_name == "lo":
                            flags = "LOOPBACK,UP,LOWER_UP"

                        self.console.print(
                            f"{index}: {iface_name}: <{flags}> mtu {mtu_value} qdisc noqueue state {state}"
                        )
                        self.console.print(
                            f"    link/ether {mac_addr} brd ff:ff:ff:ff:ff:ff"
                        )
                        index += 1

            return 0

        except Exception as e:
            self.console.print(f"[red]iplink: {e}[/red]")
            return 1



"""OtherBusybox utility commands for Cascade."""

# TODO: Find appropriate categories for these commands.

from .ar import ArCommand
from .basename import BasenameCommand
from .catv import CatvCommand
from .clear import ClearCommand
from .cpio import CpioCommand
from .cryptpw import CryptpwCommand
from .egrep import EgrepCommand
from .fgrep import FgrepCommand
from .getopt import GetoptCommand
from .ipcs import IpcsCommand
from .lzmacat import LzmacatCommand
from .nmeter import NmeterCommand
from .pipeprogress import PipeProgressCommand
from .reformine import ReformineCommand
from .reset import ResetCommand
from .sed import SedCommand
from .strings import StringsCommand
from .sysctl import SysctlCommand
from .unlzop import UnlzopCommand

__all__ = [
    "ArCommand",
    "BasenameCommand",
    "CatvCommand",
    "ClearCommand",
    "CpioCommand",
    "CryptpwCommand",
    "EgrepCommand",
    "FgrepCommand",
    "GetoptCommand",
    "IpcsCommand",
    "LzmacatCommand",
    "NmeterCommand",
    "PipeProgressCommand",
    "ReformineCommand",
    "ResetCommand",
    "SedCommand",
    "StringsCommand",
    "SysctlCommand",
    "UnlzopCommand",
]



"""Shell history management for Cascade."""

from __future__ import annotations

import logging
import os
import tempfile


class ShellHistory:
    """Manages shell command history."""

    def __init__(self, max_size: int = 1000, history_file: str | None = None):
        self.max_size = max_size
        self.history: list[str] = []
        self.current_index = 0

        if history_file:
            self.history_file = history_file
        else:
            # TODO: Use a more appropriate default path.
            self.history_file = os.path.join(tempfile.gettempdir(), ".cascade_history")

        self._load_history()

    def add_command(self, command: str) -> None:
        """Add a command to history.

        Args:
            command: Command to add to history
        """
        command = command.strip()
        if not command:
            return
        if self.history and self.history[-1] == command:
            return
        if command.startswith("history"):
            return
        self.history.append(command)
        if len(self.history) > self.max_size:
            self.history = self.history[-self.max_size :]
        self.current_index = len(self.history)
        self._save_history()

    def get_previous(self) -> str | None:
        """Get previous command in history.

        Returns:
            Previous command or None if at beginning
        """
        if not self.history or self.current_index <= 0:
            return None
        self.current_index -= 1
        return self.history[self.current_index]

    def get_next(self) -> str | None:
        """Get next command in history.

        Returns:
            Next command or None if at end
        """
        if not self.history or self.current_index >= len(self.history) - 1:
            self.current_index = len(self.history)
            return None
        self.current_index += 1
        return self.history[self.current_index]

    def get_history(self, count: int | None = None) -> list[str]:
        """Get history entries.

        Args:
            count: Number of entries to return (None for all)

        Returns:
            List of history entries
        """
        if count is None:
            return self.history.copy()
        if count <= 0:
            return []
        return self.history[-count:]

    def search_history(self, pattern: str) -> list[str]:
        """Search history for commands containing pattern.

        Args:
            pattern: Pattern to search for

        Returns:
            List of matching commands
        """
        return [cmd for cmd in self.history if pattern.lower() in cmd.lower()]

    def clear_history(self):
        """Clear all history."""
        self.history.clear()
        self.current_index = 0
        self._save_history()

    def _load_history(self):
        """Load history from file."""
        try:
            if not os.path.exists(self.history_file):
                return
            with open(self.history_file) as f:
                lines = f.readlines()
                self.history = [line.strip() for line in lines if line.strip()]
                if len(self.history) > self.max_size:
                    self.history = self.history[-self.max_size :]
                self.current_index = len(self.history)
        except Exception:
            # TODO: Add logging.
            self.history = []
            self.current_index = 0

    def _save_history(self):
        """Save history to file."""
        try:
            os.makedirs(os.path.dirname(self.history_file), exist_ok=True)
            with open(self.history_file, "w", encoding="utf-8") as f:
                for command in self.history:
                    f.write(f"{command}\n")
        except OSError as e:
            logging.warning("Failed to save history to %s: %s", self.history_file, e)

    def get_stats(self) -> dict[str, int | str | tuple[str, int] | None]:
        """Get history statistics.

        Returns:
            Dictionary with history stats
        """
        if not self.history:
            return {
                "total_commands": 0,
                "unique_commands": 0,
                "most_used": None,
                "history_file": self.history_file,
            }

        # TODO: Should be a collections.Counter
        command_counts: dict[str, int] = {}
        for cmd in self.history:
            base_cmd = cmd.split()[0] if cmd.split() else cmd
            command_counts[base_cmd] = command_counts.get(base_cmd, 0) + 1

        most_used: tuple[str, int] | None = (
            max(command_counts.items(), key=lambda x: x[1]) if command_counts else None
        )

        return {
            "total_commands": len(self.history),
            "unique_commands": len(set(self.history)),
            "most_used": most_used,
            "history_file": self.history_file,
        }

    def expand_history(self, command: str) -> str:
        """Expand history expressions in a command.

        Supports:
        - !! - Repeat last command
        - !n - Execute command number n (1-based)
        - !string - Execute last command starting with string
        - ^old^new - Quick substitution from last command

        Args:
            command: Command string that may contain history expressions

        Returns:
            Expanded command string

        Raises:
            HistoryExpansionError: If history expansion fails
        """
        if not command.strip():
            return command
        if command.startswith("^") and command.count("^") >= 2:
            return self._handle_quick_substitution(command)
        if "!" in command:
            return self._expand_bang_history(command)
        return command

    def _handle_quick_substitution(self, command: str) -> str:
        """Handle ^old^new quick substitution.

        Args:
            command: Command starting with ^ for substitution

        Returns:
            Last command with substitution applied

        Raises:
            HistoryExpansionError: If substitution fails
        """
        if not self.history:
            raise HistoryExpansionError("No previous command for substitution")

        # Parse ^old^new format
        parts = command[1:].split("^", 2)  # Skip first ^, split on remaining ^
        if len(parts) < 2:
            raise HistoryExpansionError("Invalid substitution format. Use ^old^new")

        old_text, new_text = parts
        last_command = self.history[-1]
        if old_text not in last_command:
            raise HistoryExpansionError(
                f"'{old_text}' not found in last command: {last_command}"
            )
        return last_command.replace(old_text, new_text, 1)

    def _expand_bang_history(self, command: str) -> str:
        """Expand ! history expressions in command.

        Args:
            command: Command that may contain ! expressions

        Returns:
            Command with ! expressions expanded

        Raises:
            HistoryExpansionError: If expansion fails
        """
        if not self.history:
            raise HistoryExpansionError("No command history available")
        if command == "!!":
            return self.history[-1]
        if not command.startswith("!") or len(command) <= 1:
            return command
        rest = command[1:]
        if rest.isdigit():
            cmd_num = int(rest)
            if cmd_num < 1 or cmd_num > len(self.history):
                raise HistoryExpansionError(
                    f"Command number {cmd_num} out of range (1-{len(self.history)})"
                )
            return self.history[cmd_num - 1]  # Convert to 0-based index
        for cmd in reversed(self.history):
            if cmd.startswith(rest):
                return cmd
        raise HistoryExpansionError(f"No command found starting with '{rest}'")

    def get_command_number(self, command: str) -> int | None:
        """Get the command number (1-based) for a given command.

        Args:
            command: Command to find

        Returns:
            Command number or None if not found
        """
        try:
            return self.history.index(command) + 1  # Convert to 1-based
        except ValueError:
            return None

    def get_numbered_history(
        self, start: int | None = None, count: int | None = None
    ) -> list[tuple[int, str]]:
        """Get history entries with their command numbers.

        Args:
            start: Starting command number (1-based, None for beginning)
            count: Number of entries to return (None for all remaining)

        Returns:
            List of (command_number, command) tuples
        """
        if not self.history:
            return []

        start_idx = 0 if start is None else max(0, start - 1)  # Convert to 0-based
        end_idx = (
            len(self.history)
            if count is None
            else min(len(self.history), start_idx + count)
        )

        return [
            (i + 1, cmd)
            for i, cmd in enumerate(self.history[start_idx:end_idx], start_idx)
        ]


class HistoryExpansionError(Exception):
    """Exception raised when history expansion fails."""


# Global history instance.
# TODO: This should go into the shell object.
_shell_history: ShellHistory | None = None


def get_shell_history() -> ShellHistory:
    """Get the global shell history instance."""
    global _shell_history
    if _shell_history is None:
        _shell_history = ShellHistory()
    # Type checker workaround - we know _shell_history is not None here
    return _shell_history  # type: ignore[return-value]


def init_shell_history(
    max_size: int = 1000, history_file: str | None = None
) -> ShellHistory:
    """Initialize the global shell history.

    Args:
        max_size: Maximum number of history entries
        history_file: Path to history file

    Returns:
        ShellHistory instance
    """
    global _shell_history
    _shell_history = ShellHistory(max_size, history_file)
    return _shell_history



"""Seq command for Cascade.

This module provides implementation for the seq command that generates
sequences of numbers.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags
from ...utils.theme import get_theme
from .._base import Command

if TYPE_CHECKING:
    import shimmer


# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


# TODO: Move this to an exceptions.py module.
class TextUtilsError(Exception):
    """Exception raised for text processing errors."""


class SeqCommand(Command):
    """Implementation of seq command."""

    name = "seq"
    help = "Generate sequences of numbers"
    category = "Text Utilities"

    def show_help(self):
        """Show command help."""
        help_text = """Generate sequences of numbers.

Usage: seq [OPTIONS] LAST
       seq [OPTIONS] FIRST LAST
       seq [OPTIONS] FIRST INCREMENT LAST

Options:
    -h, --help      Show this help message
    -s, --separator SEP  Use SEP instead of newline
    -w, --equal-width    Pad numbers with leading zeros
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the seq command."""
        if handle_help_flag(self, args):
            return 0

        result = parse_flags(
            args,
            {
                "f": str,  # format string
                "s": str,  # separator
                "w": bool,  # equal width (zero padding)
            },
            self.shell,
        )
        if result is None:
            return 1
        flags, positional_args = result

        format_str = flags.get("f")
        separator = flags.get("s") or "\n"  # Use "\n" if None or empty
        equal_width = flags.get("w", False)

        # Parse positional arguments
        if len(positional_args) == 1:
            # seq LAST
            first = 1.0
            increment = 1.0
            last = float(positional_args[0])
        elif len(positional_args) == 2:
            # seq FIRST LAST
            first = float(positional_args[0])
            increment = 1.0
            last = float(positional_args[1])
        elif len(positional_args) == 3:
            # seq FIRST INCREMENT LAST
            first = float(positional_args[0])
            increment = float(positional_args[1])
            last = float(positional_args[2])
        else:
            self.console.print(
                get_theme().error_text("seq: invalid number of arguments")
            )
            return 1

        if increment == 0:
            raise TextUtilsError("increment cannot be zero")

        try:
            numbers = self._generate_sequence(first, increment, last)
            if equal_width:
                # Determine the maximum width needed
                max_width = max(len(str(int(num))) for num in numbers)
                formatted_numbers = [
                    f"{int(num):0{max_width}d}" if num.is_integer() else str(num)
                    for num in numbers
                ]
            elif format_str:
                # Use custom format
                formatted_numbers = [format_str % num for num in numbers]
            else:
                # Default format
                formatted_numbers = [
                    str(int(num)) if num.is_integer() else str(num) for num in numbers
                ]

            output = separator.join(formatted_numbers)
            if separator != "\n":
                output += "\n"

            self.console.print(output, end="")
            return 0

        except (ValueError, TextUtilsError) as e:
            self.console.print(get_theme().error_text(f"seq: {e}"))
            return 1

    def _generate_sequence(
        self, first: float, increment: float, last: float
    ) -> list[float]:
        """Generate the sequence of numbers."""
        # Calculate number of steps
        if increment > 0 and first <= last:
            steps = int((last - first) / increment) + 1
        elif increment < 0 and first >= last:
            steps = int((first - last) / abs(increment)) + 1
        else:
            return []

        # Generate sequence using range-like logic
        return [first + i * increment for i in range(steps)]



"""Remote glob expansion utilities for Pebble filesystem."""

from __future__ import annotations

import fnmatch
import os
from typing import TYPE_CHECKING

import ops

if TYPE_CHECKING:
    import shimmer


def expand_remote_globs(
    client: ops.pebble.Client | shimmer.PebbleCliClient,
    pattern: str,
    base_path: str = "/",
) -> list[str]:
    """Expand glob patterns against the remote filesystem.

    Args:
        client: Pebble client for remote filesystem access
        pattern: Glob pattern to expand (e.g., "*.txt", "file*", "dir/*")
        base_path: Base directory to search from (default: "/")

    Returns:
        List of matching file paths
    """
    if not pattern or pattern == "." or pattern == "..":
        return [pattern]

    # Handle absolute paths:
    if pattern.startswith("/"):
        base_path = "/"
        pattern = pattern[1:]

    # Split pattern into directory and filename parts:
    dir_parts = pattern.split("/")
    filename_pattern = dir_parts[-1]
    dir_path = "/".join(dir_parts[:-1]) if len(dir_parts) > 1 else ""

    # Resolve the directory path:
    search_dir = os.path.join(base_path, dir_path).replace("//", "/")
    if search_dir.endswith("/") and len(search_dir) > 1:
        search_dir = search_dir[:-1]

    try:
        files = client.list_files(search_dir)
    except ops.pebble.PathError:
        return [pattern]

    matches: list[str] = []
    for file_info in files:
        filename = file_info.name

        if filename in (".", ".."):
            continue

        # Check if filename matches the pattern:
        if fnmatch.fnmatch(filename, filename_pattern):
            full_path = f"{dir_path}/{filename}" if dir_path else filename

            # Handle absolute paths:
            if pattern.startswith("/"):
                full_path = f"/{full_path}"
            elif not full_path.startswith("/"):
                full_path = os.path.join(base_path, full_path).replace("//", "/")

            matches.append(full_path)
    return sorted(matches)


def expand_remote_globs_recursive(
    client: ops.pebble.Client | shimmer.PebbleCliClient,
    pattern: str,
    base_path: str = "/",
) -> list[str]:
    """Expand glob patterns recursively against the remote filesystem.

    This function handles patterns like "**/*.txt" for recursive matching.

    Args:
        client: Pebble client for remote filesystem access
        pattern: Glob pattern to expand (supports ** for recursion)
        base_path: Base directory to search from (default: "/")

    Returns:
        List of matching file paths
    """
    if "**" not in pattern:
        return expand_remote_globs(client, pattern, base_path)

    # Split pattern by **:
    parts = pattern.split("**")
    if len(parts) != 2:
        return [pattern]

    suffix = parts[1].lstrip("/")

    matches: list[str] = []

    def search_recursive(current_dir: str, remaining_pattern: str):
        """Recursively search directories for matches."""
        try:
            files = client.list_files(current_dir)
        except ops.pebble.PathError:
            return

        for file_info in files:
            if file_info.name in (".", ".."):
                continue

            current_path = f"{current_dir}/{file_info.name}".replace("//", "/")

            if file_info.type == ops.pebble.FileType.DIRECTORY:
                # If we have a remaining pattern, continue searching:
                if remaining_pattern:
                    search_recursive(current_path, remaining_pattern)
                else:
                    # No remaining pattern, this directory matches:
                    matches.append(current_path)
                    # Continue searching recursively for nested directories:
                    search_recursive(current_path, remaining_pattern)
            # It's a file, check if it matches the remaining pattern:
            elif not remaining_pattern or fnmatch.fnmatch(
                file_info.name, remaining_pattern
            ):
                matches.append(current_path)

    # Start recursive search:
    search_recursive(base_path, suffix)

    return sorted(matches)


def expand_globs_in_tokens(
    client: ops.pebble.Client | shimmer.PebbleCliClient,
    tokens: list[str],
    base_path: str = "/",
) -> list[str]:
    """Expand glob patterns in a list of tokens.

    Args:
        client: Pebble client for remote filesystem access
        tokens: List of command tokens
        base_path: Base directory for relative paths

    Returns:
        List of tokens with glob patterns expanded
    """
    expanded: list[str] = []

    for token in tokens:
        # Check if token contains glob patterns:
        if any(char in token for char in "*?[") and not token.startswith("-"):
            matches = expand_remote_globs(client, token, base_path)
            if matches:
                expanded.extend(matches)
            else:
                # No matches, keep original token:
                expanded.append(token)
        else:
            expanded.append(token)

    return expanded



"""Implementation of CpuinfoCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

from rich.panel import Panel

from ...utils.command_helpers import handle_help_flag
from ...utils.proc_reader import ProcReadError, parse_proc_cpuinfo
from ...utils.table_builder import create_enhanced_table
from .._base import Command

if TYPE_CHECKING:
    import ops
    import shimmer
    from rich.table import Table


class CpuinfoCommand(Command):
    """Show detailed CPU information."""

    name = "cpuinfo"
    help = "Show CPU information. Use -c for compact format, -t for topology, -a for all CPUs"
    category = "System"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute cpuinfo command."""
        if handle_help_flag(self, args):
            return 0
        compact = False
        topology = False
        all_cpus = False

        for arg in args:
            if arg == "-c":
                compact = True
            elif arg == "-t":
                topology = True
            elif arg == "-a":
                all_cpus = True

        try:
            cpuinfo_data = parse_proc_cpuinfo(client)
        except ProcReadError as e:
            self.console.print(f"Error reading CPU information: {e}")
            return 1

        if compact:
            self._display_compact_cpuinfo(cpuinfo_data)
        elif topology:
            self._display_topology_cpuinfo(cpuinfo_data)
        else:
            self._display_full_cpuinfo(cpuinfo_data, all_cpus)

        return 0

    def _display_compact_cpuinfo(
        self, cpuinfo_data: dict[str, int | list[dict[str, str]]]
    ):
        """Display compact CPU information."""
        table = create_enhanced_table()
        table.add_column("CPU", style="cyan", no_wrap=True)
        table.add_column("Model", style="green", no_wrap=False)
        table.add_column("Cores", style="yellow", no_wrap=True)
        table.add_column("Cache", style="blue", no_wrap=True)

        cpus = cpuinfo_data["cpus"]
        assert isinstance(cpus, list), f"cpus should be list, got {type(cpus)}"
        for cpu_num, cpu_info in enumerate(cpus):
            self._add_cpu_row(table, cpu_info, cpu_num)

        self.console.print(table.build())

    def _add_cpu_row(self, table: Table, cpu_info: dict[str, str], cpu_num: int):
        """Add a CPU row to the table."""
        model = cpu_info.get("model name", "Unknown")
        cores = cpu_info.get("cpu cores", "1")
        cache = cpu_info.get("cache size", "Unknown")

        table.add_row(f"CPU {cpu_num}", model, cores, cache)

    def _display_topology_cpuinfo(
        self, cpuinfo_data: dict[str, int | list[dict[str, str]]]
    ):
        """Display CPU topology information."""
        table = create_enhanced_table()
        table.add_column("CPU", style="cyan", no_wrap=True)
        table.add_column("Physical ID", style="green", no_wrap=True)
        table.add_column("Core ID", style="yellow", no_wrap=True)
        table.add_column("Siblings", style="blue", no_wrap=True)
        table.add_column("Core siblings", style="magenta", no_wrap=True)

        cpus = cpuinfo_data["cpus"]
        assert isinstance(cpus, list), f"cpus should be list, got {type(cpus)}"
        for cpu_num, cpu_info in enumerate(cpus):
            self._add_topology_row(table, cpu_info, cpu_num)

        self.console.print(table.build())

    def _add_topology_row(self, table: Table, cpu_info: dict[str, str], cpu_num: int):
        """Add a topology row to the table."""
        physical_id = cpu_info.get("physical id", "0")
        core_id = cpu_info.get("core id", "0")
        siblings = cpu_info.get("siblings", "1")
        core_siblings = cpu_info.get("core siblings", "1")

        table.add_row(f"CPU {cpu_num}", physical_id, core_id, siblings, core_siblings)

    def _display_full_cpuinfo(
        self, cpuinfo_data: dict[str, int | list[dict[str, str]]], all_cpus: bool
    ):
        """Display full CPU information."""
        cpus = cpuinfo_data["cpus"]
        assert isinstance(cpus, list), f"cpus should be list, got {type(cpus)}"

        if all_cpus:
            # Show all CPUs
            content_lines = []
            for cpu_info in cpus:
                for key, value in cpu_info.items():
                    content_lines.append(f"{key}:\t{value}")
                content_lines.append("")  # Empty line between CPUs

            content = "\n".join(content_lines)
            self.console.print(
                Panel(content, title="CPU Information", border_style="bright_blue")
            )
            return

        # Show just the first CPU
        if cpus:
            first_cpu = cpus[0]
            content_lines = []
            for key, value in first_cpu.items():
                content_lines.append(f"{key}:\t{value}")

            content = "\n".join(content_lines)
            self.console.print(
                Panel(
                    content,
                    title="CPU Information",
                    border_style="bright_blue",
                )
            )



"""Integration tests for the LessCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a LessCommand instance."""
    yield pebble_shell.commands.LessCommand(shell=shell)


def test_name(command: pebble_shell.commands.LessCommand):
    assert command.name == "less"


def test_category(command: pebble_shell.commands.LessCommand):
    assert command.category == "File Utilities"


def test_help(command: pebble_shell.commands.LessCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "less" in output
    assert "View file contents with pagination" in output
    assert "-n" in output
    assert "-E" in output


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "less" in output
    assert "View file contents with pagination" in output


def test_execute_no_args_stdin(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # less with no args should read from stdin
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed reading stdin or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle stdin input
        assert len(output.strip()) >= 0
    else:
        # Should fail if no stdin available
        assert result == 1


def test_execute_single_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test viewing single file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should either succeed showing file or fail if not accessible
    if result == 0:
        output = capture.get()
        # Should show file contents
        assert len(output.strip()) > 0
        # Should contain typical /etc/passwd content patterns
        assert any(pattern in output for pattern in ["root:", "bin:", ":"])
    else:
        # Should fail if file not accessible
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["No such file", "permission denied", "error"]
        )


def test_execute_nonexistent_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with nonexistent file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/nonexistent/file.txt"])

    # Should fail with file not found error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["No such file", "not found", "error"])


def test_execute_empty_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with empty file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/dev/null"])

    # Should either succeed with empty output or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle empty file
        assert len(output.strip()) == 0
    else:
        assert result == 1


def test_execute_large_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with large file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/usr/share/dict/words"])

    # Should either succeed with pagination or fail if not found
    if result == 0:
        output = capture.get()
        # Should handle large file with pagination
        assert len(output.strip()) > 0
    else:
        # Should fail if file not found
        assert result == 1


def test_execute_binary_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with binary file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should either succeed showing binary or handle appropriately
    if result == 0:
        output = capture.get()
        # Should handle binary file (may show warning or filtered content)
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_text_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with text file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should succeed showing text content
    if result == 0:
        output = capture.get()
        # Should show text file contents
        assert len(output.strip()) > 0
        # Should contain typical hosts file patterns
        assert any(pattern in output for pattern in ["localhost", "127.0.0.1", "#"])
    else:
        assert result == 1


def test_execute_line_numbers_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test -N option for line numbers
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-N", "/etc/hosts"])

    # Should either succeed with line numbers or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show line numbers
        assert any(char.isdigit() for char in output)
    else:
        assert result == 1


def test_execute_no_line_numbers_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test -n option to suppress line numbers
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n", "/etc/hosts"])

    # Should either succeed without line numbers or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content without explicit line numbers
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_quit_at_eof_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test -E option to quit at EOF
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-E", "/etc/hosts"])

    # Should either succeed with EOF behavior or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content and quit at EOF
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_ignore_case_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test -i option for case-insensitive search
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-i", "/etc/hosts"])

    # Should either succeed with case-insensitive mode or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with case-insensitive search enabled
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_squeeze_blank_lines_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test -s option to squeeze blank lines
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "/etc/hosts"])

    # Should either succeed with squeezed blank lines or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with reduced blank lines
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_chop_long_lines_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test -S option to chop long lines
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-S", "/etc/passwd"])

    # Should either succeed with chopped lines or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with horizontal scrolling
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_raw_control_chars_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test -r option for raw control characters
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-r", "/etc/hosts"])

    # Should either succeed with raw control chars or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content with raw control characters
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_multiple_files(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with multiple files
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts", "/etc/passwd"])

    # Should either succeed showing multiple files or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show content from multiple files
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_directory_argument(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with directory argument
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc"])

    # Should either handle directory or fail appropriately
    if result == 0:
        output = capture.get()
        # Should show directory contents or error
        assert len(output.strip()) >= 0
    else:
        # Should fail if directories not supported
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["is a directory", "cannot open", "error"])


def test_execute_permission_denied_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with permission denied file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/shadow"])

    # Should fail with permission error
    if result == 1:
        output = capture.get()
        assert any(
            msg in output for msg in ["permission denied", "cannot open", "error"]
        )
    else:
        # May succeed if file is readable
        assert result == 0


def test_execute_symlink_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with symbolic link
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin"])  # Often a symlink

    # Should either follow symlink or handle appropriately
    if result == 0:
        output = capture.get()
        # Should follow symlink and show target
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_special_files_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test with special files
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/proc/version"])

    # Should either succeed with special file or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show special file contents
        assert "Linux" in output or len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_pagination_behavior(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test pagination behavior with large file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should handle pagination appropriately
    if result == 0:
        output = capture.get()
        # Should show content with pagination
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_search_functionality(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test search functionality within less
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should provide search capability
    if result == 0:
        output = capture.get()
        # Should show searchable content
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_navigation_capabilities(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test navigation capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should provide navigation features
    if result == 0:
        output = capture.get()
        # Should show navigable content
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_terminal_interaction(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test terminal interaction capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should interact with terminal appropriately
    if result == 0:
        output = capture.get()
        # Should handle terminal interaction
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_status_line_display(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test status line display
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should display status information
    if result == 0:
        output = capture.get()
        # Should include status information
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_color_support(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test color support in output
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-R", "/etc/hosts"])

    # Should handle color escape sequences
    if result == 0:
        output = capture.get()
        # Should process color information
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_unicode_support(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test unicode character support
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should handle unicode characters appropriately
    if result == 0:
        output = capture.get()
        # Should display unicode content correctly
        assert len(output.strip()) >= 0
    else:
        assert result == 1


def test_execute_tab_expansion(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test tab character expansion
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should handle tab expansion appropriately
    if result == 0:
        output = capture.get()
        # Should expand tabs correctly
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_line_wrapping(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test line wrapping behavior
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should handle line wrapping appropriately
    if result == 0:
        output = capture.get()
        # Should wrap long lines correctly
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_performance_large_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test performance with large files
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/var/log/syslog"])

    # Should handle large files efficiently
    if result == 0:
        output = capture.get()
        # Should load efficiently without hanging
        assert len(output) >= 0
    else:
        # Should fail if file not found
        assert result == 1


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should be memory efficient
    if result == 0:
        output = capture.get()
        # Should not consume excessive memory
        assert len(output) < 100000  # Reasonable output size limit
    else:
        assert result == 1


def test_execute_invalid_option_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test handling of invalid options
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-Z", "/etc/hosts"])

    # Should handle invalid options appropriately
    if result == 1:
        output = capture.get()
        # Should show error for invalid option
        assert any(msg in output for msg in ["invalid", "unknown", "option"])
    else:
        # May ignore unknown options
        assert result == 0


def test_execute_exit_behavior(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LessCommand,
):
    # Test exit behavior
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should exit appropriately
    if result == 0:
        output = capture.get()
        # Should complete and exit normally
        assert len(output.strip()) >= 0
    else:
        assert result == 1



"""Implementation of ip command (read-only)."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from ...utils.command_helpers import handle_help_flag
from ...utils.proc_reader import ProcReadError, read_proc_file
from .._base import Command

if TYPE_CHECKING:
    import ops
    import shimmer


# TODO: Use the prototype from Shimmer.
ClientType = Union["ops.pebble.Client", "shimmer.PebbleCliClient"]


class IpCommand(Command):
    """Implementation of ip command (read-only)."""

    name = "ip"
    help = "Display IP routing, network devices, policy routing and tunnels"
    category = "Network"

    def show_help(self):
        """Show command help."""
        help_text = """Display IP routing, network devices, policy routing and tunnels.

Usage: ip [OPTIONS] OBJECT [COMMAND]

Description:
    Display network configuration information. Read-only version.

Objects:
    addr, address       Protocol addresses
    route               Routing table entries
    link                Network devices
    rule                Routing policy rules

Commands:
    show, list          Show/list objects (default)

Options:
    -4, -6              Use IPv4/IPv6 only
    -h, --help          Show this help message

Examples:
    ip addr show        # Show all addresses
    ip route show       # Show routing table
    ip link show        # Show network interfaces
    ip rule show        # Show routing rules
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the ip command."""
        if handle_help_flag(self, args):
            return 0

        if not args:
            self.show_help()
            return 0

        # Parse arguments
        obj = args[0]

        try:
            if obj in ["addr", "address"]:
                return self._show_addresses(client)
            elif obj == "route":
                return self._show_routes(client)
            elif obj == "link":
                return self._show_links(client)
            elif obj == "rule":
                return self._show_rules(client)
            else:
                self.console.print(f"[red]ip: unknown object '{obj}'[/red]")
                return 1

        except Exception as e:
            self.console.print(f"[red]ip: {e}[/red]")
            return 1

    def _show_addresses(self, client: ClientType) -> int:
        """Show IP addresses."""
        # This is a simplified version - real ip addr would parse more complex data
        self.console.print("[yellow]ip addr: simplified display[/yellow]")

        try:
            # Try to read interface information
            net_dev = read_proc_file(client, "/proc/net/dev")
            interfaces = []

            for line in net_dev.split("\n")[2:]:
                if ":" in line:
                    iface_name = line.split(":")[0].strip()
                    if iface_name:
                        interfaces.append(iface_name)

            for i, iface in enumerate(interfaces, 1):
                self.console.print(
                    f"{i}: {iface}: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc noqueue state UP"
                )
                self.console.print(
                    "    link/ether 02:42:ac:11:00:02 brd ff:ff:ff:ff:ff:ff"
                )
                if iface == "lo":
                    self.console.print("    inet 127.0.0.1/8 scope host lo")
                else:
                    self.console.print(
                        f"    inet 172.17.0.2/16 brd 172.17.255.255 scope global {iface}"
                    )

            return 0

        except ProcReadError:
            self.console.print("[red]ip: cannot read network information[/red]")
            return 1

    def _show_routes(self, client: ClientType) -> int:
        """Show routing table."""
        try:
            route_info = read_proc_file(client, "/proc/net/route")
            lines = route_info.strip().split("\n")

            if len(lines) < 2:
                self.console.print("No routes found")
                return 0

            # Skip header
            for line in lines[1:]:
                fields = line.split()
                if len(fields) >= 8:
                    iface = fields[0]
                    dest = fields[1]
                    gateway = fields[2]

                    # Convert hex IP addresses to dotted decimal
                    dest_ip = self._hex_to_ip(dest)
                    gw_ip = self._hex_to_ip(gateway)

                    if dest_ip == "0.0.0.0":  # noqa: S104 # legitimate use of 0.0.0.0 for default route
                        self.console.print(f"default via {gw_ip} dev {iface}")
                    else:
                        self.console.print(f"{dest_ip} dev {iface}")

            return 0

        except ProcReadError:
            self.console.print("[red]ip: cannot read routing information[/red]")
            return 1

    def _show_links(self, client: ClientType) -> int:
        """Show network links."""
        try:
            net_dev = read_proc_file(client, "/proc/net/dev")
            lines = net_dev.strip().split("\n")

            index = 1
            for line in lines[2:]:
                if ":" in line:
                    iface_name = line.split(":")[0].strip()
                    if iface_name:
                        flags = "<BROADCAST,MULTICAST,UP,LOWER_UP>"
                        if iface_name == "lo":
                            flags = "<LOOPBACK,UP,LOWER_UP>"

                        self.console.print(
                            f"{index}: {iface_name}: {flags} mtu 1500 qdisc noqueue state UP mode DEFAULT"
                        )
                        self.console.print(
                            "    link/ether 02:42:ac:11:00:02 brd ff:ff:ff:ff:ff:ff"
                        )
                        index += 1

            return 0

        except ProcReadError:
            self.console.print(
                "[red]ip: cannot read network interface information[/red]"
            )
            return 1

    def _show_rules(self, client: ClientType) -> int:
        """Show routing policy rules."""
        # Default routing rules
        self.console.print("0:\tfrom all lookup local")
        self.console.print("32766:\tfrom all lookup main")
        self.console.print("32767:\tfrom all lookup default")
        return 0

    def _hex_to_ip(self, hex_str: str) -> str:
        """Convert hex string to IP address."""
        if len(hex_str) != 8:
            return "0.0.0.0"  # noqa: S104 # legitimate use of 0.0.0.0 as fallback IP

        try:
            # Reverse byte order for little endian
            ip_int = int(hex_str, 16)
            return f"{ip_int & 0xFF}.{(ip_int >> 8) & 0xFF}.{(ip_int >> 16) & 0xFF}.{(ip_int >> 24) & 0xFF}"
        except ValueError:
            return "0.0.0.0"  # noqa: S104 # legitimate use of 0.0.0.0 as fallback IP



"""Implementation of UnlzopCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


# TODO: Put this in the compression category.
class UnlzopCommand(Command):
    """Decompress lzop files."""

    name = "unlzop"
    help = "Decompress lzop files"
    category = "Compression"

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the unlzop command."""
        return self._execute_unlzop(client, args)

    def _execute_unlzop(self, client: ClientType, args: list[str]) -> int:
        # TODO: Implement this!
        return 1
